# -*- coding: utf-8 -*-
# pylint: disable=redefined-builtin,invalid-name,import-error
#
# gluetool documentation build configuration file, created by
# sphinx-quickstart on Sat Mar 25 14:10:42 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import os
import sys

import sphinx.ext.autodoc
import sphinx.environment
import sphinx_rtd_theme
from docutils.utils import get_source_line
from pkg_resources import get_distribution


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames as a string.
# You can specify multiple suffixes as a list of strings.
#
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'gluetool'
copyright = u'2017, mprchlik@redhat.com, mvadkert@redhat.com and others'
author = u'mprchlik@redhat.com, mvadkert@redhat.com and others'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
release = get_distribution('gluetool').version
version = '.'.join(release.split('.')[:2])  # MAJOR.MINOR is good enough

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'gluetooldoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'gluetool.tex', u'gluetool Documentation',
     u'mvadkerti@redhat.com', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'gluetool', u'gluetool Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'gluetool', u'gluetool Documentation',
     author, 'gluetool', 'One line description of project.',
     'Miscellaneous'),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}

# Disable verification of certificates - intersphinx can have some issues,
# and it's not critical
tls_verify = False

# default autodoc flags
autodoc_default_flags = ['members', 'undoc-members', 'show-inheritance', 'private-members']


# pylint: disable=unused-argument
def _warn_node(self, msg, node, *args, **kwargs):
    if not msg.startswith('nonlocal image URI found:'):
        # pylint: disable=protected-access
        self._warnfunc(msg, '%s:%s' % get_source_line(node))


sphinx.environment.BuildEnvironment.warn_node = _warn_node


# Include module description as a part of documentation
class ModuleDescriptionDocumenter(sphinx.ext.autodoc.DataDocumenter):
    objtype = "moddesc"
    content_indent = ""

    # pylint: disable=unused-argument
    def add_directive_header(self, sig):
        pass


def run_apidoc(_):
    from sphinx.apidoc import main

    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    module = os.path.join(cur_dir, '..', '..', 'gluetool')
    main([None, '-e', '-o', cur_dir, module, '--force'])


def setup(app):
    app.add_autodocumenter(ModuleDescriptionDocumenter)
    app.connect('builder-inited', run_apidoc)
